/******************************************************************************
 *  NEOPOP - Emulator as in Dreamland
 *  Copyright (C) 2001-2002 neopop_uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ****************************************************************************/

#include <neopop.h>
#include <font.h>
#include <iso9660.h>
#include <dvd.h>
#include <unzip.h>
#include <sdcard.h>

#define PAGESIZE 21
#define PADCAL 40

static int maxfiles;
int havedir = 0;
int haveSDdir = 0;
int UseSDCARD = 0;
sd_file * filehandle;
char rootSDdir[SDCARD_MAX_PATH_LEN];
int LoadDVDFile (unsigned char *buffer);
int offset = 0;
int selection = 0;
extern int neosize;
extern unsigned char *neorom;

/***************************************************************************
 * Showfile screen
 ***************************************************************************/
static void ShowFiles (int offset, int selection) 
{
  int i, j;
  char text[56];

  ClearScreen ();
  j = 0;

  for (i = offset; i < (offset + PAGESIZE) && (i < maxfiles); i++)
    {
      /* test if selection is a directory */
      if (filelist[i].flags) setfontcolour (COLOR_YELLOW);
	  else setfontcolour (COLOR_WHITE);
	  
	  memset(text,0,56);
	  strncpy (text, filelist[i].filename,55);
	  text[55] = 0;

      if (j == (selection - offset))
	{
			/*** Highlighted text entry ***/
	      fntDrawBoxFilled (100, (j * fheight) + 120, 540, ((j + 1) * fheight) + 120, COLOR_BLUE);
	  WriteCentre ((j * fheight) + 120, text);
	  }
      else WriteCentre ((j * fheight) + 120, text);
	  j++;
  }
  SetScreen ();
	}

/***************************************************************************
 * Update SDCARD curent directory name 
 ***************************************************************************/ 
int updateSDdirname()
{
  int size=0;
  char *test;
  char temp[1024];

   /* current directory doesn't change */
   if (strcmp(filelist[selection].filename,".") == 0) return 0; 
   
   /* go up to parent directory */
   else if (strcmp(filelist[selection].filename,"..") == 0) 
   {
     /* determine last subdirectory namelength */
     sprintf(temp,"%s",rootSDdir);
     test= strtok(temp,"\\");
     while (test != NULL)
     { 
       size = strlen(test);
       test = strtok(NULL,"\\");
     }
  
     /* remove last subdirectory name */
     size = strlen(rootSDdir) - size - 1;
     rootSDdir[size] = 0;

	 /* handles root name */
	 if (strcmp(rootSDdir,"dev0:") == 0) sprintf(rootSDdir,"dev0:\\NPROMS\\..");
	 
     return 1;
   }
      else
	{
     /* test new directory namelength */
     if ((strlen(rootSDdir)+1+strlen(filelist[selection].filename)) < SDCARD_MAX_PATH_LEN) 
     {
       /* handles root name */
	   if (strcmp(rootSDdir,"dev0:\\NPROMS\\..") == 0) sprintf(rootSDdir,"dev0:");
	 
	   /* update current directory name */
       sprintf(rootSDdir, "%s\\%s",rootSDdir, filelist[selection].filename);
       return 1;
     }
     else
     {
         WaitPrompt ("Dirname is too long !"); 
         return -1;
     }
    } 
	}

/***************************************************************************
 * Browse SDCARD subdirectories 
 ***************************************************************************/ 
int parseSDdirectory()
{
  int entries = 0;
  int nbfiles = 0;
  DIR *sddir = NULL;

  /* initialize selection */
  selection = offset = 0;
  
  /* Get a list of files from the actual root directory */ 
  entries = SDCARD_ReadDir (rootSDdir, &sddir);
  
  if (entries < 0) entries = 0;   
  if (entries > MAXFILES) entries = MAXFILES;
    
  /* Move to DVD structure - this is required for the file selector */ 
  while (entries)
  {
      memset (&filelist[nbfiles], 0, sizeof (FILEENTRIES));
      strncpy(filelist[nbfiles].filename,sddir[nbfiles].fname,MAXJOLIET);
      filelist[nbfiles].filename[MAXJOLIET-1] = 0;
	  filelist[nbfiles].length = sddir[nbfiles].fsize;
	  filelist[nbfiles].flags = (char)(sddir[nbfiles].fattr & SDCARD_ATTR_DIR);
      nbfiles++;
      entries--;
    }

  /*** Release memory ***/
  free(sddir);

  return nbfiles;
}

/****************************************************************************
 * FileSelector
 *
 * Let user select a file from the DVD listing
 ****************************************************************************/
void FileSelector () 
{
  short p;
  signed char a;
  int haverom = 0;
  int redraw = 1;

  while (haverom == 0)
  {
    if (redraw) ShowFiles (offset, selection);
    redraw = 0;
    p = PAD_ButtonsDown (0);
    a = PAD_StickY (0);

    /* handles User Input */
    if (p & PAD_BUTTON_B) haverom = 1;
    
    if ((p & PAD_BUTTON_DOWN) || (a < -PADCAL))
	{
	  selection++;
	  if (selection == maxfiles) selection = offset = 0;
	  if ((selection - offset) >= PAGESIZE) offset += PAGESIZE;
	  redraw = 1;
	} // End of down

    if ((p & PAD_BUTTON_UP) || (a > PADCAL))
	{
	  selection--;
	  if (selection < 0)
	  {
	      selection = maxfiles - 1;
	      offset = selection - PAGESIZE + 1;
	  }
	  if (selection < offset) offset -= PAGESIZE;
	  if (offset < 0) offset = 0;
	  redraw = 1;
	} // End of Up

    if (p & PAD_BUTTON_LEFT)
	{
	  /*** Go back a page ***/
	  selection -= PAGESIZE;
	  if (selection < 0)
	  {
	      selection = maxfiles - 1;
	      offset = selection - PAGESIZE + 1;
	  }
	  if (selection < offset) offset -= PAGESIZE;
	  if (offset < 0) offset = 0;
	  redraw = 1;
	}

    if (p & PAD_BUTTON_RIGHT)
	{
	  /*** Go forward a page ***/
	  selection += PAGESIZE;
	  if (selection > maxfiles - 1) selection = offset = 0;
	  if ((selection - offset) >= PAGESIZE) offset += PAGESIZE;
	  redraw = 1;
	}

    if (p & PAD_BUTTON_A)
	{
	  /* item selected */
	  if (filelist[selection].flags)			/*** This is directory ***/
	  {
	        if (UseSDCARD)
            {
               /* update current directory and set new entry list if directory has changed */
			   int status = updateSDdirname();
			   if (status == 1)
			   {
				   maxfiles = parseSDdirectory();
				   if (!maxfiles)
				   {
					   WaitPrompt ("Error reading directory !");
					   haverom   = 1; // quit SD menu
                       haveSDdir = 0; // reset everything at next access
			       }
			   }
			   else if (status == -1)
			   {
				   haverom   = 1; // quit SD menu
                   haveSDdir = 0; // reset everything at next access
			   }
             }
             else
             {
	           rootdir = filelist[selection].offset;
	           rootdirlength = filelist[selection].length;
	           offset = selection = 0;
	           maxfiles = parsedirectory ();
	         }
	  }
	  else
	    {
	      rootdir = filelist[selection].offset;
	      rootdirlength = filelist[selection].length;
          /*** Now load the DVD file to it's offset ***/
	      neosize = LoadDVDFile (&neorom[0]);
	      rom.length = neosize;
	      rom.data = (char *) &neorom[0];
	      rom_loaded ();
	      reset ();
	      haverom = 1;
	    }
	  redraw = 1;
	}
    }
}

/****************************************************************************
 * OpenDVD
 *
 * Function to load a DVD directory and display to user.
 ****************************************************************************/
void OpenDVD () 
{
  UseSDCARD = 0;
  while (!getpvd())
  {
	ShowAction("Mounting DVD ... Wait");
	DVD_Mount();
  }
  
  if (havedir == 0)
  {
     /* don't mess with SD entries */
	 haveSDdir = 0;
	 
	 /* reinit selector */
	 selection = offset = 0;
	 
    if ((maxfiles = parsedirectory ()))
	{
	  FileSelector ();
	  havedir = 1;
	}
  }
  else FileSelector ();
}


   
/****************************************************************************
 * OpenSD updated to use the new libogc.  Written by softdev and pasted
 * into this code by Drack.
 * Modified for subdirectory browing & quick filelist recovery
 * Enjoy!
*****************************************************************************/
int OpenSD () 
{
  UseSDCARD = 1;
  
  if (haveSDdir == 0)
  {
     /* don't mess with DVD entries */
	 havedir = 0;
	 
	 /* Initialise libOGC SD functions */ 
     SDCARD_Init ();
 
     /* Reset SDCARD root directory */
     sprintf(rootSDdir,"dev0:\\NPROMS");
 
	 /* Parse initial root directory and get entries list */
     if ((maxfiles = parseSDdirectory ()))
	 {
       /* Select an entry */
	   FileSelector ();
    
       /* memorize last entries list, actual root directory and selection for next access */
	   haveSDdir = 1;
	 }
	 else
     {
        /* no entries found */
	    WaitPrompt ("NPROMS not found on SDCARD (slot A) !");
        return 0;
    }
  }
  /* Retrieve previous entries list and made a new selection */
  else  FileSelector ();
  
  return 1;
}

/****************************************************************************
 * SDCard Get Info
 ****************************************************************************/ 
void GetSDInfo () 
{
  char fname[SDCARD_MAX_PATH_LEN];
  rootdirlength = 0;
 
  /* Check filename length */
  if ((strlen(rootSDdir)+1+strlen(filelist[selection].filename)) < SDCARD_MAX_PATH_LEN)
     sprintf(fname, "%s\\%s",rootSDdir,filelist[selection].filename); 
  
  else
  {
    WaitPrompt ("Maximum Filename Length reached !"); 
    haveSDdir = 0; // reset everything before next access
  }

  filehandle = SDCARD_OpenFile (fname, "rb");
  if (filehandle == NULL)
    {
      WaitPrompt ("Unable to open file!");
      return;
    }
  rootdirlength = SDCARD_GetFileSize (filehandle);
}


/****************************************************************************
 * LoadDVDFile
 *
 * This function will load a file from DVD, in BIN, SMD or ZIP format.
 * The values for offset and length are inherited from rootdir and 
 * rootdirlength.
 *
 * The buffer parameter should re-use the initial ROM buffer.
 ****************************************************************************/
int LoadDVDFile (unsigned char *buffer) 
{
  int offset;
  int blocks;
  int i;
  int discoffset;
  char readbuffer[2048];

  /*** SDCard Addition ***/ 
  if (UseSDCARD) GetSDInfo ();
  if (rootdirlength == 0) return 0;
  
  /*** How many 2k blocks to read ***/
  blocks = rootdirlength / 2048;
  offset = 0;
  discoffset = rootdir;
  ShowAction ("Loading ... Wait");
  
  if (UseSDCARD) SDCARD_ReadFile (filehandle, &readbuffer, 2048);
  else dvd_read (&readbuffer, 2048, discoffset);

  if (!IsZipFile ((char *) readbuffer))
  {
    if (UseSDCARD) SDCARD_SeekFile (filehandle, 0, SDCARD_SEEK_SET);
    
    for (i = 0; i < blocks; i++)
	{
	  if (UseSDCARD) SDCARD_ReadFile (filehandle, &readbuffer, 2048);
	  else dvd_read (readbuffer, 2048, discoffset);
	  memcpy (buffer + offset, readbuffer, 2048);
	  offset += 2048;
	  discoffset += 2048;
	}

	/*** And final cleanup ***/ 
	if (rootdirlength % 2048)
	{
	  i = rootdirlength % 2048;
	  if (UseSDCARD) SDCARD_ReadFile (filehandle, &readbuffer, i);
	  else dvd_read (readbuffer, 2048, discoffset);
	  memcpy (buffer + offset, readbuffer, i);
	}
  }
  else
  {
    return UnZipBuffer (buffer, discoffset, rootdirlength);
  }
  
  if (UseSDCARD) SDCARD_CloseFile (filehandle);
  return rootdirlength;
}


