/******************************************************************************
 *  NEOPOP - Emulator as in Dreamland
 *  Copyright (C) 2001-2002 neopop_uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ****************************************************************************/
#include <neopop.h>
#include <zlib.h>
#include "caps14.h"
#include "neoback.h"

extern unsigned int *xfb[2];
extern int whichfb;
extern int copynow;

/*** Backdrop ***/
char backdrop[(640 * 480 * 2) + 32];

 /*** Font support ***/
 /* Change following two lines based on font height and name */
unsigned char *pcfont = (unsigned char *) CAPS14;
int fheight = 14;
int cwidth = 8;
int forecolour = COLOR_WHITE;

/****************************************************************************
 * Set double width on or off
 ****************************************************************************/
void
setdoublewidth (int x)
{
  if (x)
    cwidth = 16;
  else
    cwidth = 8;
}

/*****************************************************************************
 * Set font colour
 *****************************************************************************/
void
setfontcolour (int fcolour)
{
  forecolour = fcolour;
}

/******************************************************************************
 * Actual Draw Character Routines
 ******************************************************************************/
void
fntDrawChar (int x, int y, unsigned char c)
{

  int offset;			// Offset in pixels
  int mask;			// Bitmask for character
  int xx, yy;			// X and Y positions
  char t;			// PC Font character
  int cpos;			// PC Font position
  int scol;			// Current colour on screen
  int cols[2];			// Used to define the on screen colour
  int j;

  /* Calculate offset on screen */
  offset = (y * 320) + (x >> 1);

  /* Get position in font array */
  cpos = c;
  //cpos <<= 4;   // Each character is 16 bits wide
  cpos = cpos * fheight;

  /* Now draw the character */
  for (yy = 0; yy < fheight; yy++)
    {
      /* Because each bit is used, the actual
         on screen colour needs to be built
         in blocks of two pixels */
      t = pcfont[cpos++];	// Get current 8-bit value
      mask = 0x80;
      j = 0;

      for (xx = 0; xx < cwidth; xx += 2)
	{

	  if (cwidth == 8)
	    {
	      scol = xfb[whichfb][offset + j];	// Get current colour twin from screen

	      if (t & mask)
		cols[0] = forecolour;
	      else
		cols[0] = scol;

	      mask >>= 1;

	      if (t & mask)
		cols[1] = forecolour;
	      else
		cols[1] = scol;

	      /* Now update the display with the current colours */
	      xfb[whichfb][offset + j++] =
		(cols[0] & 0xffff0000) | (cols[1] & 0xffff);
	    }
	  else
	    {

	      /* Now double width */
	      if (t & mask)
		xfb[whichfb][offset + j] = forecolour;

	      mask >>= 1;
	      j++;

	      if (t & mask)
		xfb[whichfb][offset + j] = forecolour;

	      j++;

	    }

	  mask >>= 1;

	}

      offset += 320;		// Move to next line

    }				// End of font row

}

/* Draw a string of characters */
void
write_font (int x, int y, char *text)
{
  int i;

  for (i = 0; i < strlen (text); i++)
    {
      fntDrawChar (x, y, text[i]);
      x += cwidth;
    }

}

/* Draw a string on the centre of the screen */
void
WriteCentre (int y, char *text)
{
  int xp;

  xp = (640 - (strlen (text) * cwidth));
  xp >>= 1;

  write_font (xp, y, text);

}

void
WriteShadow (int y, char *text, int fcolor, int shadow)
{

  int xp;
  int oldcolour;

  oldcolour = forecolour;

  xp = (640 - (strlen (text) * cwidth));
  xp >>= 1;

  forecolour = shadow;
  write_font (xp + 2, y + 2, text);
  forecolour = fcolor;
  write_font (xp, y, text);
  forecolour = oldcolour;
}

void
fntDrawHLine (int x1, int x2, int y, int color)
{
  int i;

  y = 320 * y;
  x1 >>= 1;
  x2 >>= 1;

  for (i = x1; i <= x2; i++)
    xfb[whichfb][y + i] = color;
}

void
fntDrawVLine (int x, int y1, int y2, int color)
{
  int i;

  x >>= 1;
  for (i = y1; i <= y2; i++)
    xfb[whichfb][x + (640 * i) / 2] = color;
}

void
fntDrawBox (int x1, int y1, int x2, int y2, int color)
{
  fntDrawHLine (x1, x2, y1, color);
  fntDrawHLine (x1, x2, y2, color);
  fntDrawVLine (x1, y1, y2, color);
  fntDrawVLine (x2, y1, y2, color);
}

void
fntDrawBoxFilled (int x1, int y1, int x2, int y2, int color)
{
  int h;

  for (h = y1; h <= y2; h++)
    fntDrawHLine (x1, x2, h, color);

}

void
SetScreen ()
{

  VIDEO_SetNextFramebuffer (xfb[whichfb]);
  VIDEO_Flush ();
  VIDEO_WaitVSync ();

}

void
ClearScreen ()
{
  copynow = GX_FALSE;
  whichfb ^= 1;
  memcpy (xfb[whichfb], &backdrop, 1280 * 480);
}

void
WaitPrompt (char *msg)
{
  int quit = 0;

  while (PAD_ButtonsDown (0) & PAD_BUTTON_A)
    {
    };

  while (!(PAD_ButtonsDown (0) & PAD_BUTTON_A) && (quit == 0))
    {
      ClearScreen ();
      WriteCentre (254, msg);
      WriteCentre (254 + fheight, "Press A to Continue");
      SetScreen ();

      while (!(PAD_ButtonsDown (0) & PAD_BUTTON_A));
    }
}

void
ShowAction (char *msg)
{
  ClearScreen ();
  WriteCentre (254, msg);
  SetScreen ();
}

void
WaitButtonA ()
{
  while (PAD_ButtonsDown (0) & PAD_BUTTON_A)
    {
    };
  while (!(PAD_ButtonsDown (0) & PAD_BUTTON_A));
}

/****************************************************************************
 * Unpack Backdrop
 *
 * Called at startup to unpack our backdrop to a temporary 
 * framebuffer.
 ****************************************************************************/
void
unpackBackdrop ()
{
  unsigned long res, inbytes, outbytes;

  inbytes = neoback_COMPRESSED;
  outbytes = neoback_RAW;

  res =
    uncompress ((char *) &backdrop[0], &outbytes, (char *) &neoback[0],
		inbytes);

  if (res != Z_OK)
    while (1);
}
